#include "EMA_Day.mqh"

#define MINUTE 60
#define MINUTE_MSC (MINUTE * 1000)

#define ALIGN(A, B) ((A) / (B) * (B))

class SPREADSYMBOL
{
private:
  EMA_DAY EMA_Day;

  string SymbOrigin;
  string SymbName;

  datetime From;

  static bool IsEqual( string Str1, string Str2 )
  {
    ::StringToLower(Str1);
    ::StringToLower(Str2);

    return(!::StringFind(Str1, Str2));
  }

  static string SearchSymbol( string Name )
  {
    for (int i = SymbolsTotal(false) - 1; i >= 0; i--)
    {
      const string Symb = SymbolName(i, false);

      if (SPREADSYMBOL::IsEqual(Symb, Name))
      {
        Name = Symb;

        break;
      }
    }

    return(Name);
  }

  static string GetSpreadSymbol( string Symb, string ServerName = NULL )
  {
    string Name = NULL;

    if (!::StringLen(Symb))
      Symb = _Symbol;

    const bool IsRealSymbol = !::StringLen(ServerName);

    if (IsRealSymbol)
      ServerName = ::AccountInfoString(ACCOUNT_SERVER);

    if (!::SymbolInfoInteger(Symb, SYMBOL_CUSTOM))
    {
      const string ServerName2 = ::StringSubstr(ServerName, 0, ::StringFind(ServerName, "-"));
      const string Symb2 = "SPREAD_" + ::StringSubstr(Symb, 0, 6);

      Name = SPREADSYMBOL::SearchSymbol(Symb2 + "_" + ::StringSubstr(ServerName2, 0, ::MathMin(::StringLen(ServerName2), 29 - ::StringLen(Symb2))));
    }

    return(Name);
  }

  static int CopyRates( const string &Name, const int Count, MqlRates &Rates[], int AmountTry = 5 )
  {
    int Res = 0;

    if (Count > 0)
    {
      do
      {
        ::ResetLastError();

        Res = ::CopyRates(Name, PERIOD_M1, 0, Count, Rates);
      } while (_LastError && (bool)(--AmountTry));

      if (_LastError)
        ::Print(__FILE__ + " " + (string)__LINE__ +
                ": ::CopyRates(Name, PERIOD_M1, 0, Count, Rates) = " + " = " + (string)Res +
                ", _LastError = " + (string)_LastError + ", Name = " + Name + ", Count = " + (string)Count);
    }

    return((Res > 0) ? Res : 0);
  }

  static datetime GetFrom( const string Name, double &PrevSpread )
  {
    MqlRates Rates[];

    const int Size = SPREADSYMBOL::CopyRates(Name, 1, Rates);

    PrevSpread = (Size > 0) ? Rates[Size - 1].close : -1;

    return((Size > 0) ? Rates[Size - 1].time + MINUTE : 0);
  }

  static double GetSpread( const MqlTick &Tick )
  {
    return(Tick.ask - Tick.bid);
  }

  static ulong TicksToRate( const MqlTick &Ticks[], int &Pos, double &PrevSpread, MqlRates &Rates )
  {
    ulong Res = ULONG_MAX;
    const int Size = ::ArraySize(Ticks);

    if (Pos < Size)
    {
      const ulong NextTime = ALIGN(Ticks[Pos].time_msc, MINUTE_MSC) + MINUTE_MSC;
      ulong PrevTime = (PrevSpread < 0) ? Ticks[Pos].time_msc : (NextTime - MINUTE_MSC);

      if (PrevSpread < 0)
        PrevSpread = GetSpread(Ticks[Pos]);

      Rates.high = PrevSpread;
      Rates.low = Rates.high;

      Rates.time = (datetime)((NextTime - MINUTE_MSC) / 1000);

      do
      {
        const ulong NewTime = Ticks[Pos].time_msc;
        const int Interval = (int)(((NewTime >= NextTime) ? NextTime : NewTime) - PrevTime);

        if (PrevSpread > Rates.high)
          Rates.high = PrevSpread;
        else if (PrevSpread < Rates.low)
          Rates.low = PrevSpread;

        Rates.tick_volume++;

        Rates.spread += Interval;
        Rates.open += PrevSpread * Interval;

        if (NewTime >= NextTime)
        {
          Rates.open /= Rates.spread;

          Rates.close = PrevSpread;

          Res = NewTime;

          break;
        }
        else
        {
          PrevTime = NewTime;

          PrevSpread = SPREADSYMBOL::GetSpread(Ticks[Pos++]);
        }
      } while (Pos < Size);
    }

    return(Res);
  }

  static int CalcSpreads( const string Symb, const datetime From, double PrevSpread, MqlRates &Rates[] )
  {
    MqlTick Ticks[];
    int Amount = 0;

    ::ResetLastError();
    const int Size = ::CopyTicksRange(Symb, Ticks, COPY_TICKS_INFO, From * 1000);

    if ((Size > 0) && !_LastError)
    {
      const ulong EndTime = ALIGN(Ticks[Size - 1].time_msc, MINUTE_MSC);
      const ulong BeginTime = ALIGN(Ticks[0].time_msc, MINUTE_MSC);

      if (BeginTime != EndTime)
      {
        ::ArrayResize(Rates, (int)((EndTime - BeginTime) / MINUTE_MSC));

        ::ZeroMemory(Rates);

        int Pos = 0;

        while (SPREADSYMBOL::TicksToRate(Ticks, Pos, PrevSpread, Rates[Amount++]) < EndTime)
          ;
      }
    }

    return(::ArrayResize(Rates, Amount));
  }

  void CheckExist( void )
  {
    if (this.IsReal() && !::SymbolInfoInteger(this.SymbName, SYMBOL_EXIST))
    {
      string ServerName = ::AccountInfoString(ACCOUNT_SERVER);

      ServerName = ::StringSubstr(ServerName, 0, ::StringFind(ServerName, "-"));

      ::CustomSymbolCreate(this.SymbName, "Spreads\\" + ServerName + "\\" +
                                 ::StringSubstr(::SymbolInfoString(this.SymbOrigin, SYMBOL_PATH), 0,
                                 ::StringLen(::SymbolInfoString(this.SymbOrigin, SYMBOL_PATH)) - ::StringLen(this.SymbOrigin)));

      ::CustomSymbolSetInteger(this.SymbName, SYMBOL_DIGITS, 8);
      ::CustomSymbolSetString(this.SymbName, SYMBOL_DESCRIPTION, "Created by " + __FILE__ + " (" + ::AccountInfoString(ACCOUNT_SERVER) + ")");
    }
  }

  static int ToNewRate( const MqlRates &Rates[], int Pos, MqlRates &NewRates, const int Interval )
  {
    const int Size = ::ArraySize(Rates);

    if (Pos < Size)
    {
      NewRates = Rates[Pos++];

      NewRates.open *= NewRates.spread;

      NewRates.time= ALIGN(NewRates.time, Interval);
      const datetime NextTime = NewRates.time + Interval;

      int Amount = 1;

      while ((Pos < Size) && (Rates[Pos].time < NextTime))
      {
        if (Rates[Pos].high > NewRates.high)
          NewRates.high = Rates[Pos].high;

        if (Rates[Pos].low < NewRates.low)
          NewRates.low = Rates[Pos].low;

        NewRates.open += Rates[Pos].open * Rates[Pos].spread;
        NewRates.spread += Rates[Pos].spread;

//        NewRates.close = Rates[Pos].close;
        NewRates.close += Rates[Pos].close;
        Amount++;

        NewRates.tick_volume += Rates[Pos++].tick_volume;
      }

      NewRates.open /= NewRates.spread;
      NewRates.close /= Amount;
    }

    return(Pos);
  }

public:
  SPREADSYMBOL( const int iEMA_Period = 1, const string Symb = NULL,
                const string ServerName = NULL ) : SymbOrigin(!::StringLen(ServerName) ? ((Symb == NULL) ? _Symbol : Symb) : NULL),
                                                              SymbName(SPREADSYMBOL::GetSpreadSymbol(Symb, ServerName)), From(0),
                                                              EMA_Day(iEMA_Period)
  {
  }

  void Refresh( void )
  {
    if (this.IsReal())
    {
      this.CheckExist();

      double PrevSpread;
      const datetime dFrom = ::MathMax(SPREADSYMBOL::GetFrom(this.SymbName, PrevSpread), this.From);

      MqlRates Rates[];

      const int Size = SPREADSYMBOL::CalcSpreads(this.SymbOrigin, dFrom, PrevSpread, Rates);

      if (Size)
        ::CustomRatesReplace(this.SymbName, Rates[0].time, Rates[Size - 1].time, Rates);
    }

    return;
  }

  int GetData( MqlRates &Rates[], const datetime tFrom = 0, bool Init = false, const ENUM_TIMEFRAMES ePeriod = PERIOD_CURRENT )
  {
    const int PerSeconds = ::PeriodSeconds(ePeriod);
    int Amount = 0;

    if (!tFrom)
      Init = true;

    if (PerSeconds == MINUTE)
    {
      Amount = SPREADSYMBOL::CopyRates(this.SymbName, ::Bars(this.SymbName, PERIOD_M1, ALIGN(tFrom, PerSeconds), ALIGN(::TimeCurrent(), PerSeconds)), Rates);

      if (Init)
        this.EMA_Day.Init();

      for (int i = 0; i < Amount; i++)
        Rates[i].close = this.EMA_Day.Get(Rates[i].open, Rates[i].time);
    }
    else
    {
      const int Size = this.GetData(Rates, ALIGN(tFrom, PerSeconds), Init, PERIOD_M1);

      for (int i = 0; i < Size; i = SPREADSYMBOL::ToNewRate(Rates, i, Rates[Amount++], PerSeconds))
        ;
    }

    return(::ArrayResize(Rates, Amount));
  }

  void SetFrom( const datetime tFrom )
  {
    this.From = tFrom;

    return;
  }

  bool IsReal( void ) const
  {
    return((this.SymbOrigin != NULL) && ::SymbolInfoInteger(this.SymbOrigin, SYMBOL_EXIST) &&
                                        !::SymbolInfoInteger(this.SymbOrigin, SYMBOL_CUSTOM));
  }

  string GetName( void ) const
  {
    return(this.SymbName);
  }
};

#undef ALIGN

#undef MINUTE_MSC
#undef MINUTE